/*
 * Copyright (C) 2015  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdint.h>

const static uint8_t hqTable3x_00[256] = { // table for upper left corner
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 11, 8, 1,  3, 12, 9,
  4, 4, 1, 13, 4, 4, 1, 13, 1,  3,  8, 8, 1,  3,  1, 8,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 12, 9, 1,  3, 10, 7,
  4, 4, 1, 13, 4, 4, 1, 13, 1,  3, 10, 8, 1,  3,  1, 7,
  4, 4, 1,  2, 4, 4, 1,  2, 1, 14,  8, 8, 1, 14, 10, 8,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 10, 8, 1,  3, 10, 8,
  4, 4, 1,  2, 4, 4, 1,  2, 1, 14,  1, 8, 1, 14,  1, 7,
  4, 4, 1,  2, 4, 4, 1, 13, 1,  3, 10, 8, 1, 14,  1, 7,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 11, 8, 1,  3, 12, 9,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 10, 8, 1,  3, 10, 8,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 12, 9, 1,  3, 10, 7,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 10, 9, 1,  3,  1, 7,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 10, 8, 1,  3, 10, 9,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 10, 8, 1,  3,  1, 8,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3, 10, 8, 1,  3,  1, 7,
  4, 4, 1,  2, 4, 4, 1,  2, 1,  3,  1, 8, 1,  3,  1, 7,
};

const static uint8_t hqTable3x_02[256] = { // table for upper right corner
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4, 14, 14,
  1, 1, 11, 12,  2,  2, 8, 9, 1, 1,  8,  1, 2,  2,  8,  8,
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4,  3,  3,
  1, 1, 11, 12,  2,  2, 8, 9, 1, 1, 10, 10, 2,  2,  8,  8,
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4,  3,  3,
  1, 1,  8, 10, 13, 13, 8, 8, 1, 1, 10, 10, 2,  2,  8,  8,
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4,  3,  3,
  1, 1, 10, 10,  2,  2, 8, 9, 1, 1, 10,  1, 2,  2,  8,  8,
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4, 14, 14,
  1, 1, 12, 10,  2,  2, 9, 7, 1, 1, 10,  1, 2,  2,  8,  7,
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4,  3,  3,
  1, 1, 12, 10,  2,  2, 9, 7, 1, 1, 10,  1, 2,  2,  9,  7,
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4,  3, 14,
  1, 1,  1,  1, 13, 13, 8, 7, 1, 1, 10,  1, 2, 13,  8,  7,
  4, 4,  1,  1,  4,  4, 3, 3, 4, 4,  1,  1, 4,  4,  3,  3,
  1, 1, 10,  1,  2,  2, 8, 7, 1, 1,  1,  1, 2,  2,  8,  7,
};

const static uint8_t hqTable3x_22[256] = { // table for lower right corner
   4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4, 4,
   1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 1,
   4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4, 4,
   1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 1,
   1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 1,
  11, 11,  8, 10, 12, 12,  1, 10,  8, 10, 10, 10, 10, 10, 10, 1,
   1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 1,
  12, 12, 10, 10, 10, 10,  1,  1,  1, 10, 10,  1,  1,  1,  1, 1,
   4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4, 4,
   3,  3, 14,  3,  3,  3, 14,  3,  3,  3,  3,  3,  3,  3,  3, 3,
   4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4, 4,
   3,  3, 14,  3,  3,  3, 14,  3,  3,  3,  3,  3,  3,  3, 14, 3,
   2,  2,  2,  2,  2,  2,  2,  2, 13,  2,  2,  2, 13,  2,  2, 2,
   8,  8,  8,  8,  9,  9,  8,  8,  8,  8,  8,  8,  8,  9,  8, 8,
   2,  2,  2,  2,  2,  2,  2,  2, 13,  2,  2,  2, 13,  2, 13, 2,
   9,  9,  8,  9,  7,  7,  7,  7,  8,  8,  8,  8,  7,  7,  7, 7,
};

const static uint8_t hqTable3x_20[256] = { // table for lower left corner
   4,  4, 4,  4, 4, 4, 4, 4,  1,  1,  1,  1,  1,  1,  1,  1,
   4,  4, 4,  4, 4, 4, 4, 4,  1,  1,  1,  1,  1,  1,  1,  1,
   4,  4, 4,  4, 4, 4, 4, 4,  2,  2, 13, 13,  2,  2,  2,  2,
   4,  4, 4,  4, 4, 4, 4, 4,  2,  2,  2,  2,  2,  2,  2,  2,
   1,  1, 1,  1, 1, 1, 1, 1, 11, 12,  8,  1, 11, 12, 10, 10,
   1,  1, 1,  1, 1, 1, 1, 1,  8, 10, 10, 10, 10, 10, 10,  1,
   3,  3, 3,  3, 3, 3, 3, 3,  8,  9,  8,  8,  8,  9,  8,  8,
  14, 14, 3,  3, 3, 3, 3, 3,  8,  8,  8,  8,  8,  9,  8,  8,
   4,  4, 4,  4, 4, 4, 4, 4,  1,  1,  1,  1,  1,  1,  1,  1,
   4,  4, 4,  4, 4, 4, 4, 4,  1,  1,  1,  1,  1,  1,  1,  1,
   4,  4, 4,  4, 4, 4, 4, 4,  2,  2, 13, 13,  2,  2,  2,  2,
   4,  4, 4,  4, 4, 4, 4, 4,  2,  2,  2, 13,  2,  2,  2,  2,
   1,  1, 1,  1, 1, 1, 1, 1, 12, 10, 10,  1, 12, 10, 10,  1,
   1,  1, 1,  1, 1, 1, 1, 1,  1,  1, 10,  1, 10,  1,  1,  1,
   3,  3, 3,  3, 3, 3, 3, 3,  9,  7,  8,  7,  9,  7,  9,  7,
  14, 14, 3, 14, 3, 3, 3, 3,  8,  7,  8,  7,  8,  7,  8,  7,
};

const static uint8_t hqTable3x_01[256] = { // table for upper middle corner
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 8, 8, 1, 1, 9, 9,
  1, 1, 5, 6, 1, 1, 5, 6, 1, 1, 0, 8, 1, 1, 5, 0,
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 7, 7, 1, 1, 0, 0,
  1, 1, 5, 6, 1, 1, 5, 6, 1, 1, 0, 8, 1, 1, 5, 0,
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 8, 8, 1, 1, 0, 8,
  1, 1, 5, 0, 1, 1, 5, 5, 1, 1, 0, 8, 1, 1, 5, 0,
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 8, 1, 1, 0, 0,
  1, 1, 0, 0, 1, 1, 5, 6, 1, 1, 0, 8, 1, 1, 5, 8,
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 8, 8, 1, 1, 9, 9,
  1, 1, 4, 0, 1, 1, 4, 0, 1, 1, 0, 8, 1, 1, 5, 0,
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 7, 7, 1, 1, 0, 0,
  1, 1, 4, 0, 1, 1, 4, 0, 1, 1, 0, 7, 1, 1, 4, 0,
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 8, 1, 1, 0, 9,
  1, 1, 0, 0, 1, 1, 5, 0, 1, 1, 0, 8, 1, 1, 5, 5,
  1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 8, 1, 1, 0, 0,
  1, 1, 0, 0, 1, 1, 5, 0, 1, 1, 0, 8, 1, 1, 5, 0,
};

const static uint8_t hqTable3x_12[256] = { // table for right middle corner
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 8, 7, 0, 0, 8, 7, 0, 0, 8, 0, 0, 0, 8, 8,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 8, 7, 0, 0, 8, 7, 0, 0, 0, 0, 0, 0, 8, 8,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  5, 5, 0, 0, 6, 6, 8, 8, 5, 0, 0, 0, 0, 0, 8, 8,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  4, 4, 0, 0, 0, 0, 8, 7, 0, 0, 0, 0, 0, 0, 8, 8,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 9, 0, 0, 0, 9, 0, 0, 0, 0, 0, 0, 0, 8, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 9, 0, 0, 0, 9, 0, 0, 0, 0, 0, 0, 0, 9, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  5, 5, 5, 5, 6, 6, 0, 0, 5, 5, 5, 5, 5, 6, 0, 8,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  4, 4, 5, 4, 0, 0, 0, 0, 5, 5, 5, 5, 0, 0, 5, 0,
};

const static uint8_t hqTable3x_21[256] = { // table for lower middle corner
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 5, 4, 5, 0, 5, 4, 0, 0,
  8, 8, 8, 0, 7, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 5, 4, 5, 5, 5, 4, 5, 5,
  9, 9, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 4, 5, 5,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 6, 0, 0, 0,
  8, 8, 8, 8, 7, 7, 8, 8, 8, 8, 8, 8, 8, 7, 8, 8,
  0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 5, 0, 6, 0, 6, 0,
  9, 9, 8, 9, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 8, 0,
};

const static uint8_t hqTable3x_10[256] = { // table for left middle corner
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 5, 5, 0, 0, 4, 4,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 5, 5, 0, 0, 0, 5,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 6, 6, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 5, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 8, 9, 0, 5, 8, 9, 0, 5,
  1, 1, 1, 1, 1, 1, 1, 1, 8, 0, 0, 5, 0, 0, 0, 5,
  1, 1, 1, 1, 1, 1, 1, 1, 8, 9, 8, 0, 8, 9, 8, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 8, 8, 8, 0, 8, 9, 8, 5,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 5, 5, 0, 0, 4, 4,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 5, 0, 0, 0, 5,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 6, 6, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 6, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 7, 0, 0, 5, 7, 0, 0, 4,
  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 5, 0, 0, 0, 5,
  1, 1, 1, 1, 1, 1, 1, 1, 7, 0, 8, 0, 7, 0, 7, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 8, 0, 8, 8, 8, 0, 8, 0,
};


void hq3x_16_Init(uint32_t *hqTable) {
    int index;
    uint8_t *_hqTable8;

    _hqTable8 = (uint8_t *) hqTable;
    for (index = 0; index < 256; index++) {
        _hqTable8[4*index+0] = hqTable3x_00[index] | (hqTable3x_02[index] << 4);
        _hqTable8[4*index+1] = hqTable3x_22[index] | (hqTable3x_20[index] << 4);
        _hqTable8[4*index+2] = hqTable3x_01[index] | (hqTable3x_12[index] << 4);
        _hqTable8[4*index+3] = hqTable3x_21[index] | (hqTable3x_10[index] << 4);
    }
}

