/*
 * Copyright (C) 2015  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <libc64_dsp/include/inc_overlay.h>
#include "scalers.h"
//#include <stdint.h>
#include "hqx-common.h"
#include "hqx-interp.h"

static inline void prepare_hqtable2(uint8_t * restrict hqTable2)
{
    hqTable2[ 0] =  4; hqTable2[ 1] = 0;
    hqTable2[ 2] = 10; hqTable2[ 3] = 0;
    hqTable2[ 4] = 11; hqTable2[ 5] = 0;
    hqTable2[ 6] =  4; hqTable2[ 7] = 1;
    hqTable2[ 8] =  9; hqTable2[ 9] = 1;
    hqTable2[10] = 10; hqTable2[11] = 1;
}

static void fill_blendtable(const uint8_t * restrict patternTable, const uint32_t * restrict hqTable, const uint8_t * restrict hqTable2, uint16_t * restrict blendTable, int width)
{
    _nassert((int)patternTable % 8 == 0);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)blendTable % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int index, hqindex;
    uint16_t * restrict blendValues;

    for (index = 0; index < 80; index++)
    {
        blendTable[index] = 0xffff;
    }

    blendValues = (uint16_t *) (blendTable+80);

    _nassert((int)blendValues % 8 == 0);

    // fill blendtables based on patterntable

    for (index = 0; index < width; index++)
    {
        uint32_t hqValue, rule;

        hqValue = hqTable[patternTable[index]];

        rule = hqValue & 0x1f;
        blendValues[index] = blendTable[rule];
        blendTable[rule] = index;

        rule = (hqValue >> 8) & 0x1f;
        blendValues[index + MAX_WIDTH] = blendTable[rule + 20];
        blendTable[rule + 20] = index;

        rule = (hqValue >> 16) & 0x1f;
        blendValues[index + 2*MAX_WIDTH] = blendTable[rule + 2*20];
        blendTable[rule + 2*20] = index;

        rule = (hqValue >> 24) & 0x1f;
        blendValues[index + 3*MAX_WIDTH] = blendTable[rule + 3*20];
        blendTable[rule + 3*20] = index;
    }

    // modify blendtables based on further comparisons

    // upper left pixel

#define blendExpr12 patternTable[index-1] & (1 << 2)
#define blendExpr18 patternTable[index+1] & (1 << 0)
#define blendExpr19 patternTable[index-1] & (1 << 7)

#include "hq2x_fill_blendtable_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    // upper right pixel

#define blendExpr12 patternTable[index+1] & (1 << 0)
#define blendExpr18 patternTable[index+1] & (1 << 5)
#define blendExpr19 patternTable[index-1] & (1 << 2)

#include "hq2x_fill_blendtable_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    // lower right pixel

#define blendExpr12 patternTable[index+1] & (1 << 5)
#define blendExpr18 patternTable[index-1] & (1 << 7)
#define blendExpr19 patternTable[index+1] & (1 << 0)

#include "hq2x_fill_blendtable_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    // lower left pixel

#define blendExpr12 patternTable[index-1] & (1 << 7)
#define blendExpr18 patternTable[index-1] & (1 << 2)
#define blendExpr19 patternTable[index+1] & (1 << 5)

#include "hq2x_fill_blendtable_proc_body.h"
}

static void fill_dst(const uint16_t * restrict rgbsrc1, const uint16_t * restrict rgbsrc2, const uint16_t * restrict rgbsrc3, uint16_t * restrict dst, int width, const uint16_t * restrict blendTable)
{
    _nassert((int)rgbsrc1 % 8 == 0);
    _nassert((int)rgbsrc2 % 8 == 0);
    _nassert((int)rgbsrc3 % 8 == 0);
    _nassert((int)blendTable % 8 == 0);

    int index;
    const uint16_t * restrict blendValues;

    blendValues = (uint16_t *) (blendTable+80);

    _nassert((int)blendValues % 8 == 0);


    _nassert((int)dst % 8 == 0);
    // upper left pixel

#define E rgbsrc2[index]
#define A rgbsrc1[index-1]
#define B rgbsrc1[index]
#define D rgbsrc2[index-1]
#define F rgbsrc2[index+1]
#define H rgbsrc3[index]

#include "hq2x_fill_dst_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    dst += 1;

    _nassert((int)dst % 2 == 0);
    _nassert((int)(dst-1) % 8 == 0);
    // upper right pixel

#define E rgbsrc2[index]
#define A rgbsrc1[index+1]
#define B rgbsrc2[index+1]
#define D rgbsrc1[index]
#define F rgbsrc3[index]
#define H rgbsrc2[index-1]

#include "hq2x_fill_dst_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    dst += 2*width;

    _nassert((int)dst % 2 == 0);
    _nassert((int)(dst-1) % 8 == 0);
    // lower right pixel

#define E rgbsrc2[index]
#define A rgbsrc3[index+1]
#define B rgbsrc3[index]
#define D rgbsrc2[index+1]
#define F rgbsrc2[index-1]
#define H rgbsrc1[index]

#include "hq2x_fill_dst_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    dst -= 1;

    _nassert((int)dst % 8 == 0);
    // lower left pixel

#define E rgbsrc2[index]
#define A rgbsrc3[index-1]
#define B rgbsrc2[index-1]
#define D rgbsrc3[index]
#define F rgbsrc1[index]
#define H rgbsrc2[index+1]

#include "hq2x_fill_dst_proc_body.h"
}

void _hq2x_8_16(const uint8_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height, uint32_t *hqTable, uint16_t *rgbpalette)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);
    _nassert(height >= 4);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)rgbpalette % 8 == 0);

    // hqTable and rgbpalette should be in L1D SRAM

    uint8_t * restrict tmpsrc3;
    uint8_t * restrict tmpsrc2;
    uint8_t * restrict tmpsrc1;
    uint8_t * restrict tmpsrc0;

    uint16_t * restrict rgbsrc3;
    uint16_t * restrict rgbsrc2;
    uint16_t * restrict rgbsrc1;

    uint32_t * restrict yuvsrc3;
    uint32_t * restrict yuvsrc2;
    uint32_t * restrict yuvsrc1;

    uint8_t * restrict patternsrc1;

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    uint8_t * restrict hqTable2;

    uint64_t * restrict rgbyuvpalette;

    uint16_t * restrict blendTable;

    src += (height - 1) * srcstride; // src now points to last line

    tmpsrc3 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP); // tmpsrc3 is aligned to 8

    // start copying last line to tmpsrc3
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src, width);

    dst += (height - 1) * dststride; // dst now points to last two lines

    tmpsrc2 = (uint8_t *) ( ((uintptr_t) tmpsrc3) + width); // tmpsrc2 is aligned to 8
    tmpsrc1 = (uint8_t *) ( ((uintptr_t) tmpsrc2) + width); // tmpsrc1 is aligned to 8
    tmpsrc0 = (uint8_t *) ( ((uintptr_t) tmpsrc1) + width); // tmpsrc0 is aligned to 8

    rgbsrc3 = (uint16_t *) ( ((uintptr_t) tmpsrc0) +     width + 8); // rgbsrc3 is aligned to 8 and has place for rgbsrc3[-1] and rgbsrc3[width]
    rgbsrc2 = (uint16_t *) ( ((uintptr_t) rgbsrc3) + 2 * width + 8); // rgbsrc2 is aligned to 8 and has place for rgbsrc2[-1] and rgbsrc2[width]
    rgbsrc1 = (uint16_t *) ( ((uintptr_t) rgbsrc2) + 2 * width + 8); // rgbsrc1 is aligned to 8 and has place for rgbsrc1[-1] and rgbsrc1[width]

    yuvsrc3 = (uint32_t *) ( ((uintptr_t) rgbsrc1) + 2 * width + 8); // yuvsrc3 is aligned to 8 and has place for yuvsrc3[-1] and yuvsrc3[width]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc3) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc1 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * width + 8); // yuvsrc1 is aligned to 8 and has place for yuvsrc1[-1] and yuvsrc1[width]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc1) + 4 * width + 8); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[width]

    tmpdst1 = (uint16_t *) ( ((uintptr_t) patternsrc1) + width + 8); // tmpdst1 is aligned to 8
    tmpdst2 = (uint16_t *) ( ((uintptr_t) tmpdst1    ) + 8 * width); // tmpdst2 is aligned to 8

    hqTable2 = (uint8_t *) ( ((uintptr_t) tmpdst2) + 8 * width); // hqTable2 is aligned to 8

    rgbyuvpalette = (uint64_t *) ( ((uintptr_t) hqTable2) + 16); // rgbyuvpalette is aligned to 8

    blendTable = (uint16_t *) ( ((uintptr_t) rgbyuvpalette) + 8 * 256); // blendTable is aligned to 8

    prepare_hqtable2(hqTable2);

    prepare_yuv_palette_rgb_16(rgbpalette, rgbyuvpalette);

    // wait for copying last line to tmpsrc3
    syscalls.qdma_wait(QDMA_0);

    // start copying last-1 line to tmpsrc2
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src - srcstride, width);
    src -= 2*srcstride; // src now points to last-2 line

    // convert last line to yuv and rgb
    convert_yuv_8_rgb_16(tmpsrc3, yuvsrc3, rgbsrc3, rgbyuvpalette, width);

    // wait for copying last-1 line to tmpsrc2
    syscalls.qdma_wait(QDMA_0);

    // start copying last-2 line to tmpsrc1
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, width);
    src -= srcstride; // src now points to last-3 line

    // convert last-1 line to yuv and rgb
    convert_yuv_8_rgb_16(tmpsrc2, yuvsrc2, rgbsrc2, rgbyuvpalette, width);

    // calculate pattern for last line
    calculate_pattern(yuvsrc2, yuvsrc3, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(rgbsrc2, rgbsrc3, rgbsrc3, tmpdst1, width, blendTable);

    // start copying last two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    dst -= dststride; // dst now points to last-1 two lines

    for (height -= 3; height != 0; height--)
    {
        // wait for copying "last-2" line to tmpsrc1
        syscalls.qdma_wait(QDMA_0);

        // convert "last-2" line to yuv and rgb
        convert_yuv_8_rgb_16(tmpsrc1, yuvsrc1, rgbsrc1, rgbyuvpalette, width);

        // calculate pattern for "last-1" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

        // calculate and process blend tables
        fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

        // start copying "last-3" line to tmpsrc0
        syscalls.qdma_copy1d(QDMA_0, tmpsrc0, src, width);
        src -= srcstride; // src now points to last-4 line

        fill_dst(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2, width, blendTable);

        // wait for copying "last" two destination lines from tmpdst1
        syscalls.qdma_wait(QDMA_1);

        // start copying "last-1" two destination lines from tmpdst2
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        dst -= dststride; // dst now points to last-2 two lines

        // "rename" buffers
        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint8_t * restrict tmpsrc;
            tmpsrc = tmpsrc3;
            tmpsrc3 = tmpsrc2;
            tmpsrc2 = tmpsrc1;
            tmpsrc1 = tmpsrc0;
            tmpsrc0 = tmpsrc;
        }
        {
            uint16_t * restrict rgbsrc;
            rgbsrc = rgbsrc3;
            rgbsrc3 = rgbsrc2;
            rgbsrc2 = rgbsrc1;
            rgbsrc1 = rgbsrc;
        }
        {
            uint32_t * restrict yuvsrc;
            yuvsrc = yuvsrc3;
            yuvsrc3 = yuvsrc2;
            yuvsrc2 = yuvsrc1;
            yuvsrc1 = yuvsrc;
        }
    }

    // second line:

    // wait for copying first line to tmpsrc1
    syscalls.qdma_wait(QDMA_0);

    // convert first line to yuv and rgb
    convert_yuv_8_rgb_16(tmpsrc1, yuvsrc1, rgbsrc1, rgbyuvpalette, width);

    // calculate pattern for second line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2, width, blendTable);

    // wait for copying third two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_1);

    // start copying second two destination lines from tmpdst2
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    dst -= dststride; // dst now points to first two lines

    // first line:

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(rgbsrc1, rgbsrc1, rgbsrc2, tmpdst1, width, blendTable);

    // start copying second two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    //dst -= dststride;

    // wait for copying second two destination lines from tmpdst2
    syscalls.qdma_wait(QDMA_1);

    // wait for copying first two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_0);
}

void _hq2x_16_16(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height, uint32_t *hqTable)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);
    _nassert(height >= 4);
    _nassert((int)hqTable % 8 == 0);

    // hqTable should be in L1D SRAM

    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc0;

    uint32_t * restrict yuvsrc3;
    uint32_t * restrict yuvsrc2;
    uint32_t * restrict yuvsrc1;

    uint8_t * restrict patternsrc1;

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    uint8_t * restrict hqTable2;

    uint16_t * restrict blendTable;

    src += (height - 1) * srcstride / 2; // src now points to last line

    tmpsrc3 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8); // tmpsrc3 is aligned to 8 and has place for tmpsrc3[-1] and tmpsrc3[width]

    // start copying last line to tmpsrc3
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src, 2*width);

    dst += (height - 1) * dststride; // dst now points to last two lines

    tmpsrc2 = (uint16_t *) ( ((uintptr_t) tmpsrc3) + 2 * width + 8); // tmpsrc2 is aligned to 8 and has place for tmpsrc2[-1] and tmpsrc2[width]
    tmpsrc1 = (uint16_t *) ( ((uintptr_t) tmpsrc2) + 2 * width + 8); // tmpsrc1 is aligned to 8 and has place for tmpsrc1[-1] and tmpsrc1[width]
    tmpsrc0 = (uint16_t *) ( ((uintptr_t) tmpsrc1) + 2 * width + 8); // tmpsrc0 is aligned to 8 and has place for tmpsrc0[-1] and tmpsrc0[width]

    yuvsrc3 = (uint32_t *) ( ((uintptr_t) tmpsrc0) + 2 * width + 8); // yuvsrc3 is aligned to 8 and has place for yuvsrc3[-1] and yuvsrc3[width]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc3) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc1 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * width + 8); // yuvsrc1 is aligned to 8 and has place for yuvsrc1[-1] and yuvsrc1[width]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc1) + 4 * width + 8); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[width]

    tmpdst1 = (uint16_t *) ( ((uintptr_t) patternsrc1) + width + 8); // tmpdst1 is aligned to 8
    tmpdst2 = (uint16_t *) ( ((uintptr_t) tmpdst1    ) + 8 * width); // tmpdst2 is aligned to 8

    hqTable2 = (uint8_t *) ( ((uintptr_t) tmpdst2) + 8 * width); // hqTable2 is aligned to 8

    blendTable = (uint16_t *) ( ((uintptr_t) hqTable2) + 16); // blendTable is aligned to 8

    prepare_hqtable2(hqTable2);

    // wait for copying last line to tmpsrc3
    syscalls.qdma_wait(QDMA_0);

    // start copying last-1 line to tmpsrc2
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src - srcstride/2, 2*width);
    src -= srcstride; // src now points to last-2 line

    // convert last line to yuv
    convert_yuv_16(tmpsrc3, yuvsrc3, width);

    // wait for copying last-1 line to tmpsrc2
    syscalls.qdma_wait(QDMA_0);

    // start copying last-2 line to tmpsrc1
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, 2*width);
    src -= srcstride/2; // src now points to last-3 line

    // convert last-1 line to yuv
    convert_yuv_16(tmpsrc2, yuvsrc2, width);

    // calculate pattern for last line
    calculate_pattern(yuvsrc2, yuvsrc3, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst1, width, blendTable);

    // start copying last two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    dst -= dststride; // dst now points to last-1 two lines

    for (height -= 3; height != 0; height--)
    {
        // wait for copying "last-2" line to tmpsrc1
        syscalls.qdma_wait(QDMA_0);

        // convert "last-2" line to yuv
        convert_yuv_16(tmpsrc1, yuvsrc1, width);

        // calculate pattern for "last-1" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

        // calculate and process blend tables
        fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

        // start copying "last-3" line to tmpsrc0
        syscalls.qdma_copy1d(QDMA_0, tmpsrc0, src, 2*width);
        src -= srcstride/2; // src now points to "last-4" line

        fill_dst(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2, width, blendTable);

        // wait for copying "last" two destination lines from tmpdst1
        syscalls.qdma_wait(QDMA_1);

        // start copying "last-1" two destination lines from tmpdst2
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        dst -= dststride; // dst now points to "last-2" two lines

        // "rename" buffers
        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc3;
            tmpsrc3 = tmpsrc2;
            tmpsrc2 = tmpsrc1;
            tmpsrc1 = tmpsrc0;
            tmpsrc0 = tmpsrc;
        }
        {
            uint32_t * restrict yuvsrc;
            yuvsrc = yuvsrc3;
            yuvsrc3 = yuvsrc2;
            yuvsrc2 = yuvsrc1;
            yuvsrc1 = yuvsrc;
        }
    }

    // second line:

    // wait for copying first line to tmpsrc1
    syscalls.qdma_wait(QDMA_0);

    // convert first line to yuv
    convert_yuv_16(tmpsrc1, yuvsrc1, width);

    // calculate pattern for second line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2, width, blendTable);

    // wait for copying third two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_1);

    // start copying second two destination lines from tmpdst2
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    dst -= dststride; // dst now points to first two lines

    // first line:

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst1, width, blendTable);

    // start copying first two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    //dst -= dststride;

    // wait for copying second two destination lines from tmpdst2
    syscalls.qdma_wait(QDMA_1);

    // wait for copying first two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_0);
}

